class Robot:
    """
    Class representing a single "Toy Robot".  Robots instances have knowledge
    of their position, direction, and movement limits (i.e. "board" size).

    Note: Direction is specified via compass points (north, west, etc).  As per
    the original problem description, the coordinates (0, 0) represent the
    SOUTHWEST corner of the board.

    Attributes:
        DEFAULT_MAX_X (int): The default maximum allowable horizontal
            coordinate (inclusive).
        DEFAULT_MAX_Y (int): The default maximum allowable vertical coordinate
            (inclusive).
        DIRECTIONS (dict): A dictionary (str: int) of direction names (e.g.
            NORTH, EAST etc) and their numerical encoding.  Only the keys are
            intended for use externally as a list of valid directions.
        VALID_COMMANDS (list): A list of valid commands that can be interpreted
            by the robot.  The PLACE command requires at least two additional
            parameters.  See the complete problem description for more details.
    """

    DEFAULT_MAX_X = 5
    DEFAULT_MAX_Y = 5

    # Directions ordered such that adding 1 corresponds to RIGHT turn
    DIRECTIONS = {
        "NORTH": 0,
        "NORTHEAST": 1,
        "EAST": 2,
        "SOUTHEAST": 3,
        "SOUTH": 4,
        "SOUTHWEST": 5,
        "WEST": 6,
        "NORTHWEST": 7
    }

    VALID_COMMANDS = [
        "BLOCK",
        "PLACE",
        "MOVE",
        "LEFT",
        "RIGHT",
        "REPORT"
    ]

    # Private internals
    # Key corresponds to numerical direction defined in Robot.directions
    # Value is an offset vector in with x and y keys (positive x is NORTH,
    # positive y is EAST)
    _MOVEMENT_VECTORS = {
        0: {"x": 0, "y": 1},
        1: {"x": 1, "y": 1},
        2: {"x": 1, "y": 0},
        3: {"x": 1, "y": -1},
        4: {"x": 0, "y": -1},
        5: {"x": -1, "y": -1},
        6: {"x": -1, "y": 0},
        7: {"x": -1, "y": 1},
    }

    def __init__(self, max_x: int = DEFAULT_MAX_X, max_y: int = DEFAULT_MAX_Y):
        """
        Creates a new Robot instance with the specified position limits
        (optional).  Implicitly, the minimum limit on coordinates is 0 both
        horizontally and vertically.

        See the class attributes `DEFAULT_MAX_X` and `DEFAULT_MAX_Y` for the
        respective default numerical values.

        Args:
            max_x (int): The maximum allowable horizontal coordinate
                (inclusive, positive is EAST).
            max_y (int): The maximum allowable vertical coordinate
                (inclusive, positive is NORTH).

        Raises:
            ValueError: If max_x or max_y are less than 0.
        """

        if max_x < 0 or max_y < 0:
            raise ValueError("Cannot specify negative limits")

        self._max_x = max_x
        self._max_y = max_y

        self._position_x = None
        self._position_y = None
        self._direction = None

        self._blocks = []

    def valid_position(self, x: int, y: int) -> bool:
        """
        Calculates whether the given coordinates are valid for the limits set
        on this Robot instance.

        This function is used by the `move()` function to verify that a move
        action would be successful.

        Args:
            x: Proposed horizontal coordinate.
            y: Proposed vertical coordinate.

        Returns:
            bool: True if given coordinates are within limits, otherwise False.
        """

        within_limits = x >= 0 and y >= 0 and x <= self._max_x and y <= self._max_y
        on_block = (x, y) in self._blocks
        # print(self._blocks)
        # print(within_limits)
        # print(on_block)
        return within_limits and not on_block

    def add_block(self, x: int, y: int):
        if x == self._position_x and y == self._position_y:
            return

        self._blocks.append((x, y))

    def is_initialized(self):
        """
        Returns whether this Robot instance has been initialized (i.e. whether
        a valid `place()` command has been executed).

        This function is used by the `move()` function to verify that the
        instance has been correctly placed before movement.

        Returns:
            bool: True if correctly initialized, otherwise false.
        """
        return (
            self._position_x is not None
            and self._position_y is not None
            and self._direction is not None
        )

    def move(self):
        """
        Moves the robot one space in the direction it is currently facing,
        provided that said movement would be within limits.

        Will do nothing (fail silently) if this Robot instance has not been
        initialized successfully with `place()` or the specified movement would
        be out of bounds.
        """

        if not self.is_initialized():
            return

        vector = Robot._MOVEMENT_VECTORS[self._direction]
        new_position_x = self._position_x + vector["x"]
        new_position_y = self._position_y + vector["y"]

        if not self.valid_position(new_position_x, new_position_y):
            return

        self._position_x = new_position_x
        self._position_y = new_position_y

    def place(self, position_x: int, position_y: int, 
              direction_name: str = None):
        """
        Places the Robot instance at the specified coordinates with the
        specified direction.

        Will do nothing if the given coordinates are out of bounds, or the
        direction is not one of the keys in `Robot.DIRECTIONS`.

        See `__init__()` for detailed coordinate information.

        Args:
            position_x (int): Horizontal coordinate for placement.
            position_y (int): Vertical coordinate for placement.
            direction_name (str): Direction of placement; must be one of the
                string keys in `Robot.DIRECTIONS`.  Must be specified on first
                placement (or this call will fail silently), but is optional
                thereafter.
        """

        # Must be careful not to make any state changes until all inputs
        # have been validated
        # print('Checking position...')
        if not self.valid_position(position_x, position_y):
            return
        # print('Position check passed.')

        if direction_name is not None:
            direction_name = direction_name.upper()
            if direction_name not in Robot.DIRECTIONS.keys():
                return
        elif not self.is_initialized():
            # Direction MUST be specified on first placement
            return

        if direction_name is not None:
            self._direction = Robot.DIRECTIONS[direction_name]
        self._position_x = position_x
        self._position_y = position_y

    def get_position(self) -> (int, int):
        """
        Returns the current position of this Robot instance.

        Returns:
            (int, int): A tuple with this instance's current coordinates in the
                order (position_x, position_y), or (None, None) if this
                instance has not yet been initialized.
        """

        return (
            (self._position_x, self._position_y)
            if self.is_initialized()
            else (None, None)
        )

    def get_direction(self) -> str:
        """
        Returns the direction in which this instance is currently facing in
        string form.

        Returns:
            str: The current direction of this instance; will be one of the
                keys of `Robot.DIRECTION`.  Will return None if this instance
                has not yet been initialized.
        """

        return (
            {v: k for k, v in Robot.DIRECTIONS.items()}[self._direction]
            if self.is_initialized()
            else None
        )

    def rotate_left(self):
        """
        Rotates this Robot instance's direction to the LEFT.  For example, if
        currently facing NORTH, the new direction will be WEST.

        Does nothing if this instance has not yet been correctly placed
        (initialized).
        """

        if not self.is_initialized():
            return

        self._direction = (self._direction - 1) % len(Robot.DIRECTIONS)

    def rotate_right(self):
        """
        Rotates this Robot instance's direction to the RIGHT.  For example, if
        currently facing NORTH, the new direction will be EAST.

        Does nothing if this instance has not yet been correctly placed
        (initialized).
        """

        if not self.is_initialized():
            return

        self._direction = (self._direction + 1) % len(Robot.DIRECTIONS)

    def interpret_command(self, command: str):
        """
        Interprets a given string command and applies the appropriate 
        transformation to this Robot instance.  Fails silently if the command
        is unrecognized or invalid.

        Refer to the full problem description for a list and examples of valid
        commands.

        Args:
            command (str): The command to be interpreted.
        """

        command = command.upper()
        command_tokens = [x.strip() for x in command.split(' ') if len(x) > 0]

        # This is not strictly necessary as case
        if len(command_tokens) == 0 or not command_tokens[0] in \
                Robot.VALID_COMMANDS:
            return

        match command_tokens[0]:
            case "BLOCK":
                # print('Found block...')
                try:
                    if len(command_tokens) < 2:
                        # print('Insufficient tokens...')
                        return

                    parameter_tokens = [x.strip() for x in command_tokens[1].split(',')]

                    if len(parameter_tokens) < 2:
                        # print('Insufficient parameters...')
                        return
                    block_x = int(parameter_tokens[0])
                    block_y = int(parameter_tokens[1])

                    # print("Adding block: {} {}".format(block_x, block_y))
                    self.add_block(block_x, block_y)
                except ValueError as ve:
                    # print('Integer parsing error...')
                    return
            case "PLACE":
                try:
                    # Must have parameters
                    if len(command_tokens) < 2:
                        return
                    parameter_tokens = [x.strip() for x in \
                        command_tokens[1].split(',')]

                    # Must have at least X, Y
                    if len(parameter_tokens) < 2:
                        return

                    # Throws ValueError if invalid input
                    place_x = int(parameter_tokens[0])
                    place_y = int(parameter_tokens[1])

                    # Direction parameter is optional on second and subsequent 
                    # placements.  The place() method accounts for an absent
                    # direction on first call and fails silently.
                    # print("PLACE command parsed: {}, {}".format(place_x, place_y))
                    if len(parameter_tokens) > 2:
                        place_direction = parameter_tokens[2]
                        self.place(place_x, place_y, place_direction)
                    else:
                        self.place(place_x, place_y)
                except ValueError as ve:
                    # Unable to convert x or y token to int
                    return
            case "MOVE":
                self.move()
            case "LEFT":
                self.rotate_left()
            case "RIGHT":
                self.rotate_right()
            case "REPORT":
                print("Output: {}".format(str(self)))

    def __str__(self) -> str:
        """
        Returns a string representation of the instance including position and
        direction.

        Returns:
            str: This instance's string representation.
        """

        if not self.is_initialized():
            return "Uninitialized"

        return "{},{},{} | BLOCKS : {}".format(
            self._position_x,
            self._position_y,
            self.get_direction(),
            ":".join([str(x) for x in self._blocks])
        )
